/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file is part of the interface (glue layer) between an Embedded Wizard
*   generated UI application and the board support package (BSP) of a dedicated
*   target.
*   Please note: The implementation of this module is partially based on
*   examples that are provided within the GD32 Demo Suite. In case you want
*   to adapt this module to your custom specific hardware, please adapt the
*   hardware initialization code according your needs.
*   This template is responsible to initialize the system clock of the hardware,
*   to provide a timer tick for the Embedded Wizard UI application and to get
*   access to the realtime clock (RTC). Additionally, this file contains some
*   performance measurements to analyse the CPU usage of the UI application.
*
*******************************************************************************/

#include "ewconfig.h"
#include "gd32f4xx.h"
#include "ew_bsp_clock.h"

#include <string.h>

/* Defines to select adaquate clock source */
#undef RTC_CLOCK_SOURCE_IRC32K
#define RTC_CLOCK_SOURCE_LXTAL

#ifdef RTC_CLOCK_SOURCE_IRC32K
  #define RTC_ASYNCH_PREDIV   0x63
  #define RTC_SYNCH_PREDIV    0x013F;
#endif

#ifdef RTC_CLOCK_SOURCE_LXTAL
  #define RTC_ASYNCH_PREDIV   0x7F
  #define RTC_SYNCH_PREDIV    0x00FF
#endif


#define RtcTicksPerSecond    ( 1UL )
#define RtcTicksPerMinute    ( RtcTicksPerSecond * 60UL )
#define RtcTicksPerHour      ( RtcTicksPerMinute * 60UL )
#define RtcTicksPerDay       ( RtcTicksPerHour   * 24UL )

const unsigned long DaysToMonth[] =
{
  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365
};

const unsigned long DaysToMonthInLeapYear[] =
{
  0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366
};


#define ValueToBcd( aValue ) (( aValue / 10 ) << 4 + ( aValue % 10 ))
#define BcdToValue( aValue ) (( aBcd >> 4 ) * 10   + ( aBcd & 0x0F ))

static unsigned char valueToBcd( unsigned char aValue )
{
  unsigned char bcd;

  bcd  = ( aValue / 10 ) << 4;
  bcd += ( aValue % 10 );

  return bcd;
}

static unsigned char bcdToValue( unsigned char aBcd )
{
  unsigned char value;

  value  = ( aBcd >> 4 ) * 10;
  value += ( aBcd & 0x0F );

  return value;
}




#if EW_CPU_LOAD_MEASURING == 1

  /* variables for cycle counter / CPU load measurement */
  static unsigned long        WorkingTime  = 0;
  static unsigned long        SleepingTime = 0;
  static unsigned long        PrevCycCnt   = 0;

  /* helper function to initialize measuring unit (cycle counter) */
  static void CycleCounterInit( void )
  {
    /* Enable TRC */
    CoreDebug->DEMCR |= CoreDebug_DEMCR_TRCENA_Msk;

    /* Unlock DWT registers */
    if ((*(uint32_t*)0xE0001FB4) & 1)
      *(uint32_t*)0xE0001FB0 = 0xC5ACCE55;

    /* clear the cycle counter */
    DWT->CYCCNT = 0;

    /* start the cycle counter */
    DWT->CTRL = 0x40000001;

    /* initialize counters */
    WorkingTime   = 0;
    SleepingTime  = 0;
    PrevCycCnt    = 0;
  }

  /* helper function to get cycle counter value since last call */
  static unsigned long GetCycleCounterDelta( void )
  {
    unsigned long cycCnt;
    unsigned long result;

    /* get delta between current counter value and last counter value */
    cycCnt = DWT->CYCCNT;
    result = cycCnt - PrevCycCnt;
    PrevCycCnt = cycCnt;

    return result;
  }

#endif


/*******************************************************************************
* FUNCTION:
*   EwBspClockInit
*
* DESCRIPTION:
*   Initialises the system clock and the real time clock.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspClockInit( void )
{
  /* enable PMU clock */
  rcu_periph_clock_enable(RCU_PMU);
  /* enable the access of the RTC registers */
  pmu_backup_write_enable();

  #if defined (RTC_CLOCK_SOURCE_IRC32K)
        rcu_osci_on(RCU_IRC32K);
        rcu_osci_stab_wait(RCU_IRC32K);
        rcu_rtc_clock_config(RCU_RTCSRC_IRC32K);
  #elif defined (RTC_CLOCK_SOURCE_LXTAL)
        rcu_osci_on(RCU_LXTAL);
        rcu_osci_stab_wait(RCU_LXTAL);
        rcu_rtc_clock_config(RCU_RTCSRC_LXTAL);
  #else
  #error RTC clock source should be defined.
  #endif /* RTC_CLOCK_SOURCE_IRC32K */

  rcu_periph_clock_enable(RCU_RTC);
  rtc_register_sync_wait();
}


/*******************************************************************************
* FUNCTION:
*   EwBspClockGetTime
*
* DESCRIPTION:
*   Returns the current time in seconds since 01.01.1970.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   The current time in seconds since 01.01.1970.
*
*******************************************************************************/
unsigned long EwBspClockGetTime( void )
{
  rtc_parameter_struct rtcParams;

  rtc_current_time_get( &rtcParams );

  unsigned long   days;
  unsigned long   year   = bcdToValue( rtcParams.year );
  unsigned long   month  = bcdToValue( rtcParams.month );
  unsigned long   day    = bcdToValue( rtcParams.date );
  unsigned long   hour   = bcdToValue( rtcParams.hour );
  unsigned long   minute = bcdToValue( rtcParams.minute );
  unsigned long   second = bcdToValue( rtcParams.second );
  unsigned long   timeInSeconds;

  /* year is the number of years since 2000 */
  /* calculate number of days since 01.01.1970 until begin of current
     year including the additional days of leap years */

  days = (year + 30) * 365 + (year + 27) / 4;

  if (year & 0x3)
  {
    days += DaysToMonth[month - 1];
  }
  else
  {
    days += DaysToMonthInLeapYear[month - 1];
  }

  days += day;

  timeInSeconds =
    days   * RtcTicksPerDay +
    hour   * RtcTicksPerHour +
    minute * RtcTicksPerMinute +
    second;

  return timeInSeconds;
}


/*******************************************************************************
* FUNCTION:
*   EwBspClockSetTime
*
* DESCRIPTION:
*   Sets the given time in seconds since 01.01.1970 at real time clock (RTC).
*
* ARGUMENTS:
*   aTime - the time in seconds since 01.01.1970 to set in real time clock.
*
* RETURN VALUE:
*   None.
*
*******************************************************************************/
void EwBspClockSetTime( unsigned long aTime )
{
  #define LeapYearsFromBeginToNext400Year   7
  #define DaysPer4Years       ( 4 * 365 + 1 )

  rtc_parameter_struct rtcParams;

  unsigned int year;
  unsigned int year4;
  unsigned int month;
  unsigned int day;
  unsigned int hour;
  unsigned int minute;
  unsigned int second;

  hour    = ( unsigned int ) (( aTime / RtcTicksPerHour    ) % 24 );
  minute  = ( unsigned int ) (( aTime / RtcTicksPerMinute  ) % 60 );
  second  = ( unsigned int ) (( aTime / RtcTicksPerSecond  ) % 60 );

  /* calculation of the date */
  day       = ( unsigned int ) ( aTime / RtcTicksPerDay + 365);

  if ( day >= 48213 )
    day++;

  year4     = day / DaysPer4Years;
  day      -= year4 * DaysPer4Years;
  year      = day / 365;

  if ( year == 4 )
  {
    year = 3;

    if ( !( year4 % 100 == LeapYearsFromBeginToNext400Year )
       && ( year4 %  25 == LeapYearsFromBeginToNext400Year ))
      year = 4;
  }

  day  -= year * 365;
  year  = year4 * 4 + year + 1969;

  /* all months have less than 32 days, days >> 5 is a good estimate
     for the month. */

  month = day >> 5;

  /* correct month if necessary */
  if ( ( year % 4 != 0 ) ||
      (( year % 100 == 0 ) && ( year % 400 != 0 )) )
  {
    /* no leap year */
    if ( day >= DaysToMonth[ month + 1 ] )
      month++;

    day -= DaysToMonth[ month ];
  }
  else
  {
    /* leap year */
    if ( day >= DaysToMonthInLeapYear[ month + 1 ] )
      month++;

    day -= DaysToMonthInLeapYear[ month ];
  }

  memset( &rtcParams, 0, sizeof( rtcParams ));

  rtcParams.factor_asyn    = RTC_ASYNCH_PREDIV;
  rtcParams.factor_syn     = RTC_SYNCH_PREDIV;
  rtcParams.year           = valueToBcd( year - 2000 );
  rtcParams.month          = valueToBcd( month + 1 );
  rtcParams.date           = valueToBcd( day + 1 );
  rtcParams.hour           = valueToBcd( hour );
  rtcParams.minute         = valueToBcd( minute );
  rtcParams.second         = valueToBcd( second );
  rtcParams.day_of_week    = 0;
  rtcParams.display_format = RTC_24HOUR;
  rtcParams.am_pm          = RTC_AM;

  rtc_init( &rtcParams );
}


/*******************************************************************************
* FUNCTION:
*   EwBspClockGetCpuLoad
*
* DESCRIPTION:
*   Returns the current CPU load as percent value.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   The current CPU load.
*
*******************************************************************************/
 int EwBspClockGetCpuLoad( void )
{
  int cpuLoad = 0;

  #if EW_CPU_LOAD_MEASURING == 1

    /* get actual counter value */
    WorkingTime += GetCycleCounterDelta();

    /* calculate CPU load as percent value */
    cpuLoad = ( int ) ( WorkingTime / (( SleepingTime + WorkingTime ) / 100 ));

    /* clear accumulated counter values */
    WorkingTime  = 0;
    SleepingTime = 0;

  #endif

  return cpuLoad;
}


/*******************************************************************************
* FUNCTION:
*   EwBspClockCpuLoadSetActive
*
* DESCRIPTION:
*   Starts the CPU load counting. Call this function whenever CPU processing
*   is needed.
*
* ARGUMENTS:
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspClockCpuLoadSetActive( void )
{
  #if EW_CPU_LOAD_MEASURING == 1

    static char initialized = 0;

    if ( !initialized )
    {
      CycleCounterInit();
      initialized = 1;
    }

    SleepingTime += GetCycleCounterDelta();

  #endif
}


/*******************************************************************************
* FUNCTION:
*   EwBspClockCpuLoadSetIdle
*
* DESCRIPTION:
*   Stops the CPU load counting. Call this function whenever CPU processing is
*   currently not needed since the program execution is waiting for some system
*   event.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspClockCpuLoadSetIdle( void )
{
  #if EW_CPU_LOAD_MEASURING == 1

    WorkingTime += GetCycleCounterDelta();

  #endif
}


/* msy */
