/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file is part of the interface (glue layer) between an Embedded Wizard
*   generated UI application and the board support package (BSP) of a dedicated
*   target.
*   Please note: The implementation of this module is partially based on
*   examples that are provided within the GD32 Demo Suite. In case you want
*   to adapt this module to your custom specific hardware, please adapt the
*   hardware initialization code according your needs.
*   This template is responsible to initialize the display hardware of the board
*   and to provide the necessary access to update the display content.
*   The color format of the framebuffer has to correspond to the color format
*   of the Graphics Engine.
*
*   Important: This file is intended to be used as a template. Please adapt the
*   implementation according your particular hardware.
*
*******************************************************************************/

#include "ewconfig.h"
#include "gd32f4xx.h"
#include <string.h>
#include "ewrte.h"
#include "ew_bsp_display.h"
#include "ew_bsp_clock.h"
#include "ew_bsp_os.h"

#if EW_USE_OPERATING_SYSTEM == 1

  static XSemaphoreHandle     LcdUpdateSemaphore = 0;

#endif

#define LAYER_INDEX           0

#if EW_USE_DOUBLE_BUFFER == 1

  static volatile uint32_t    CurrentFramebuffer = 0;
  static volatile uint32_t    PendingFramebuffer = 0;

#else

  static volatile int         CurrentLine = -1;
  static volatile int         PendingLine = -1;

#endif

#ifndef EW_FRAME_BUFFER_ADDR
unsigned long frameBuffer[EW_FRAME_BUFFER_SIZE / sizeof(unsigned long)];

#define EW_FRAME_BUFFER_ADDR frameBuffer
#endif


#define HORIZONTAL_SYNCHRONOUS_PULSE  41
#define HORIZONTAL_BACK_PORCH         2
#define ACTIVE_WIDTH                  480
#define HORIZONTAL_FRONT_PORCH        2

#define VERTICAL_SYNCHRONOUS_PULSE    10
#define VERTICAL_BACK_PORCH           2
#define ACTIVE_HEIGHT                 272
#define VERTICAL_FRONT_PORCH          2


/* Helper function to sync on a certain line number updated by the display.
   This function is necessary when the system is used in single framebuffer mode. */
#if EW_USE_DOUBLE_BUFFER == 0
static void SyncOnLine( int aLine )
{
  /* add the vsync area and the vertical back porch */
  aLine = aLine + VERTICAL_SYNCHRONOUS_PULSE + VERTICAL_BACK_PORCH + 1;

  /* ensure that requested line is still inside active area or vertical front porch */
  if ( aLine > ACTIVE_HEIGHT + VERTICAL_FRONT_PORCH + 1 )
    aLine = 0;

  CurrentLine = -1;
  PendingLine = aLine;

  CPU_LOAD_SET_IDLE();

  /* program line number of next field */
  tli_line_mark_set( aLine );
  tli_reload_config(TLI_REQUEST_RELOAD_EN);

  #if EW_USE_OPERATING_SYSTEM == 1

    EwBspOsSemaphoreWait( LcdUpdateSemaphore, 1000 );

  #else

    while( aLine != CurrentLine )
      ;

  #endif

  CPU_LOAD_SET_ACTIVE();
}
#endif



/*!
    \brief      configure TLI GPIO
    \param[in]  none
    \param[out] none
    \retval     none
*/
static void tli_gpio_config(void)
{
  /* GPIO clock enable */
  rcu_periph_clock_enable(RCU_GPIOB);
  rcu_periph_clock_enable(RCU_GPIOE);
  rcu_periph_clock_enable(RCU_GPIOH);
  rcu_periph_clock_enable(RCU_GPIOI);
  rcu_periph_clock_enable(RCU_GPIOG);
  rcu_periph_clock_enable(RCU_GPIOF);

  /* TLI pins AF configure */
  gpio_af_set(GPIOE,GPIO_AF_14,GPIO_PIN_5);
  gpio_af_set(GPIOE,GPIO_AF_14,GPIO_PIN_6);
  gpio_af_set(GPIOE,GPIO_AF_14,GPIO_PIN_4);

  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_2);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_3);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_8);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_9);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_10);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_11);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_12);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_13);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_14);
  gpio_af_set(GPIOH,GPIO_AF_14,GPIO_PIN_15);

  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_0);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_1);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_2);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_4);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_5);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_6);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_7);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_9);
  gpio_af_set(GPIOI,GPIO_AF_14,GPIO_PIN_10);

  gpio_af_set(GPIOG,GPIO_AF_14,GPIO_PIN_6);
  gpio_af_set(GPIOG,GPIO_AF_14,GPIO_PIN_7);
  gpio_af_set(GPIOG,GPIO_AF_14,GPIO_PIN_10);
  gpio_af_set(GPIOG,GPIO_AF_14,GPIO_PIN_11);
  gpio_af_set(GPIOG,GPIO_AF_14,GPIO_PIN_12);

  gpio_af_set(GPIOF,GPIO_AF_14,GPIO_PIN_10);

  /* configure HSYNC(PI10), VSYNC(PI9), PCLK(PG7) */
  /* configure LCD_R7(PG6), LCD_R6(PH12), LCD_R5(PH11), LCD_R4(PH10), LCD_R3(PH9),LCD_R2(PH8),
               LCD_R1(PH3), LCD_R0(PH2), LCD_G7(PI2), LCD_G6(PI1), LCD_G5(PI0), LCD_G4(PH15),
               LCD_G3(PH14), LCD_G2(PH13),LCD_G1(PE6), LCD_G0(PE5),LCD_B7(PI7), LCD_B6(PI6),
               LCD_B5(PI5), LCD_B4(PI4), LCD_B3(PG11),LCD_B2(PG10), LCD_B1(PG12), LCD_B0(PE4) */
  gpio_mode_set(GPIOE, GPIO_MODE_AF, GPIO_PUPD_NONE, GPIO_PIN_4|GPIO_PIN_5|GPIO_PIN_6);
  gpio_output_options_set(GPIOE, GPIO_OTYPE_PP, GPIO_OSPEED_50MHZ,GPIO_PIN_4|GPIO_PIN_5|GPIO_PIN_6);

  gpio_mode_set(GPIOH, GPIO_MODE_AF, GPIO_PUPD_NONE, GPIO_PIN_2|GPIO_PIN_3|GPIO_PIN_8|GPIO_PIN_9|GPIO_PIN_10
               |GPIO_PIN_11|GPIO_PIN_12|GPIO_PIN_13|GPIO_PIN_14|GPIO_PIN_15);
  gpio_output_options_set(GPIOH, GPIO_OTYPE_PP, GPIO_OSPEED_50MHZ,GPIO_PIN_2|GPIO_PIN_3|GPIO_PIN_8|GPIO_PIN_9|GPIO_PIN_10
                         |GPIO_PIN_11|GPIO_PIN_12|GPIO_PIN_13|GPIO_PIN_14|GPIO_PIN_15);

  gpio_mode_set(GPIOI, GPIO_MODE_AF, GPIO_PUPD_NONE, GPIO_PIN_0|GPIO_PIN_1|GPIO_PIN_2
               |GPIO_PIN_4|GPIO_PIN_5|GPIO_PIN_6|GPIO_PIN_7|GPIO_PIN_9|GPIO_PIN_10);
  gpio_output_options_set(GPIOI, GPIO_OTYPE_PP, GPIO_OSPEED_50MHZ,GPIO_PIN_0|GPIO_PIN_1|GPIO_PIN_2
                         |GPIO_PIN_4|GPIO_PIN_5|GPIO_PIN_6|GPIO_PIN_7|GPIO_PIN_9|GPIO_PIN_10);

  gpio_mode_set(GPIOG, GPIO_MODE_AF, GPIO_PUPD_NONE, GPIO_PIN_6|GPIO_PIN_7|GPIO_PIN_10|GPIO_PIN_11|GPIO_PIN_12);
  gpio_output_options_set(GPIOG, GPIO_OTYPE_PP,GPIO_OSPEED_50MHZ, GPIO_PIN_6|GPIO_PIN_7|GPIO_PIN_10|GPIO_PIN_11|GPIO_PIN_12);

  gpio_mode_set(GPIOF, GPIO_MODE_AF, GPIO_PUPD_NONE, GPIO_PIN_10);
  gpio_output_options_set(GPIOF, GPIO_OTYPE_PP,GPIO_OSPEED_50MHZ, GPIO_PIN_10);

  /* configure LCD_PWM_BackLight(PB15) */
  gpio_mode_set(GPIOB, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, GPIO_PIN_15);
  gpio_output_options_set(GPIOB, GPIO_OTYPE_PP,GPIO_OSPEED_50MHZ, GPIO_PIN_15);

  gpio_bit_set(GPIOB,GPIO_PIN_15);
}


/*!
    \brief      configure TLI peripheral
    \param[in]  none
    \param[out] none
    \retval     none
*/
static void tli_config(void)
{
  tli_parameter_struct               tli_init_struct;
  tli_layer_parameter_struct         tli_layer_init_struct;

  rcu_periph_clock_enable(RCU_TLI);
  tli_gpio_config();

  /* configure the PLLSAI clock to generate lcd clock */
  if(ERROR == rcu_pllsai_config(192 + 21, 2, 3)) {
      while(1);
  }

  rcu_tli_clock_div_config(RCU_PLLSAIR_DIV8);
  rcu_osci_on(RCU_PLLSAI_CK);
  if(ERROR == rcu_osci_stab_wait(RCU_PLLSAI_CK)) {
      while(1);
  }

  /* configure TLI parameter struct */
  tli_init_struct.signalpolarity_hs = TLI_HSYN_ACTLIVE_LOW;
  tli_init_struct.signalpolarity_vs = TLI_VSYN_ACTLIVE_LOW;
  tli_init_struct.signalpolarity_de = TLI_DE_ACTLIVE_LOW;
  tli_init_struct.signalpolarity_pixelck = TLI_PIXEL_CLOCK_TLI;

  /* LCD display timing configuration */
  tli_init_struct.synpsz_hpsz = HORIZONTAL_SYNCHRONOUS_PULSE - 1;
  tli_init_struct.synpsz_vpsz = VERTICAL_SYNCHRONOUS_PULSE - 1;
  tli_init_struct.backpsz_hbpsz = HORIZONTAL_SYNCHRONOUS_PULSE + HORIZONTAL_BACK_PORCH - 1;
  tli_init_struct.backpsz_vbpsz = VERTICAL_SYNCHRONOUS_PULSE + VERTICAL_BACK_PORCH - 1;
  tli_init_struct.activesz_hasz = HORIZONTAL_SYNCHRONOUS_PULSE + HORIZONTAL_BACK_PORCH + ACTIVE_WIDTH - 1;
  tli_init_struct.activesz_vasz = VERTICAL_SYNCHRONOUS_PULSE + VERTICAL_BACK_PORCH + ACTIVE_HEIGHT - 1;
  tli_init_struct.totalsz_htsz = HORIZONTAL_SYNCHRONOUS_PULSE + HORIZONTAL_BACK_PORCH + ACTIVE_WIDTH + HORIZONTAL_FRONT_PORCH - 1;
  tli_init_struct.totalsz_vtsz = VERTICAL_SYNCHRONOUS_PULSE + VERTICAL_BACK_PORCH + ACTIVE_HEIGHT + VERTICAL_FRONT_PORCH - 1;
  /* configure LCD background R,G,B values */
  tli_init_struct.backcolor_red = 0xFF;
  tli_init_struct.backcolor_green = 0;
  tli_init_struct.backcolor_blue = 0;
  tli_init(&tli_init_struct);

  /* TLI layer configuration */
  tli_layer_struct_para_init( &tli_layer_init_struct );
  tli_layer_init_struct.layer_window_leftpos = HORIZONTAL_SYNCHRONOUS_PULSE + HORIZONTAL_BACK_PORCH;
  tli_layer_init_struct.layer_window_rightpos = tli_layer_init_struct.layer_window_leftpos + EW_DISPLAY_WIDTH - 1;
  tli_layer_init_struct.layer_window_toppos = VERTICAL_SYNCHRONOUS_PULSE + VERTICAL_BACK_PORCH;
  tli_layer_init_struct.layer_window_bottompos = tli_layer_init_struct.layer_window_toppos + EW_DISPLAY_HEIGHT - 1;

  /* Pixel Format configuration: translate framebuffer color format into LTDC mode */
  #if ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_RGBA8888 )
    tli_layer_init_struct.layer_ppf = LAYER_PPF_ARGB8888;
  #elif ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_RGB888 )
    tli_layer_init_struct.layer_ppf = LAYER_PPF_RGB888;
  #elif ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_RGBA4444 )
    tli_layer_init_struct.layer_ppf = LAYER_PPF_ARGB4444;
  #elif ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_RGB565 )
    tli_layer_init_struct.layer_ppf = LAYER_PPF_RGB565;
  #elif ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_Index8 )
    tli_layer_init_struct.layer_ppf = LAYER_PPF_L8;
  #elif ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_LumA44 )
    tli_layer_init_struct.layer_ppf = LAYER_PPF_AL44;
  #endif

  tli_layer_init_struct.layer_sa = 0xFF;
  tli_layer_init_struct.layer_default_blue = 0xFF;
  tli_layer_init_struct.layer_default_green = 0xFF;
  tli_layer_init_struct.layer_default_red = 0xFF;
  tli_layer_init_struct.layer_default_alpha = 0x0;
  tli_layer_init_struct.layer_acf1 = LAYER_ACF1_PASA;
  tli_layer_init_struct.layer_acf2 = LAYER_ACF2_PASA;
  tli_layer_init_struct.layer_frame_bufaddr = (uint32_t)EW_FRAME_BUFFER_ADDR;
  tli_layer_init_struct.layer_frame_line_length = EW_DISPLAY_WIDTH * EW_FRAME_BUFFER_DEPTH;
  tli_layer_init_struct.layer_frame_buf_stride_offset = EW_DISPLAY_WIDTH * EW_FRAME_BUFFER_DEPTH;
  tli_layer_init_struct.layer_frame_total_line_number = EW_DISPLAY_HEIGHT;
  tli_layer_init(LAYER0, &tli_layer_init_struct);

  tli_line_mark_set( 0 );

  tli_interrupt_enable( TLI_INT_LM );
  NVIC_EnableIRQ( TLI_IRQn );
  NVIC_SetPriority( TLI_IRQn, 0x4U );
}


void TLI_IRQHandler(void)
{
  if ( tli_flag_get(TLI_FLAG_LM) == SET )
  {
    tli_interrupt_flag_clear( TLI_FLAG_LM );

    #if EW_USE_DOUBLE_BUFFER == 1

      register uint32_t pendingBuffer = PendingFramebuffer;
      if ( CurrentFramebuffer != pendingBuffer )
      {
        /* change address within layer configuration structure */
        TLI_LxFBADDR(LAYER0) = pendingBuffer;
        tli_reload_config(TLI_REQUEST_RELOAD_EN);

        #if EW_USE_OPERATING_SYSTEM == 1

          EwBspOsSemaphoreRelease( LcdUpdateSemaphore );

        #else

          /* save new address */
          CurrentFramebuffer = pendingBuffer;

        #endif
      }

      /* program next V-sync */
      tli_line_mark_set( EW_DISPLAY_HEIGHT + VERTICAL_SYNCHRONOUS_PULSE + VERTICAL_BACK_PORCH - 1 );

    #else

      /* update current line number */
      CurrentLine = PendingLine;

      #if EW_USE_OPERATING_SYSTEM == 1

        EwBspOsSemaphoreRelease( LcdUpdateSemaphore );

      #endif

    #endif
  }
}



/*******************************************************************************
* FUNCTION:
*   EwBspDisplayInit
*
* DESCRIPTION:
*   The function EwBspDisplayInit initializes the display hardware and returns
*   the display parameter.
*
* ARGUMENTS:
*   aGuiWidth,
*   aGuiHeight   - Size of the GUI in pixel.
*   aDisplayInfo - Display info data structure.
*
* RETURN VALUE:
*   Returns 1 if successful, 0 otherwise.
*
*******************************************************************************/
int EwBspDisplayInit( int aGuiWidth, int aGuiHeight, XDisplayInfo* aDisplayInfo )
{
  EW_UNUSED_ARG( aGuiWidth );
  EW_UNUSED_ARG( aGuiHeight );

  /* check and clean display info structure */
  if ( !aDisplayInfo )
    return 0;

  memset( aDisplayInfo, 0, sizeof( XDisplayInfo ));

  /* initialize given framebuffer */
  EwZero((void*)EW_FRAME_BUFFER_ADDR, EW_FRAME_BUFFER_WIDTH * EW_FRAME_BUFFER_HEIGHT * EW_FRAME_BUFFER_DEPTH );

  #if EW_USE_OPERATING_SYSTEM == 1

    /* create the LCD update semaphore */
    LcdUpdateSemaphore = EwBspOsSemaphoreCreate( 1, 1 );

  #endif

  /* configure TLI peripheral */
  tli_config();

  /* enable TLI layers */
  tli_layer_enable(LAYER0);
  tli_enable();

  tli_reload_config(TLI_REQUEST_RELOAD_EN);

  /* return the current display configuration */
  aDisplayInfo->FrameBuffer    = (void*)EW_FRAME_BUFFER_ADDR;
  aDisplayInfo->DoubleBuffer   = (void*)EW_DOUBLE_BUFFER_ADDR;
  aDisplayInfo->BufferWidth    = EW_FRAME_BUFFER_WIDTH;
  aDisplayInfo->BufferHeight   = EW_FRAME_BUFFER_HEIGHT;
  aDisplayInfo->DisplayWidth   = EW_DISPLAY_WIDTH;
  aDisplayInfo->DisplayHeight  = EW_DISPLAY_HEIGHT;

  #if EW_USE_DOUBLE_BUFFER == 1
    aDisplayInfo->UpdateMode   = EW_BSP_DISPLAY_UPDATE_NORMAL;
  #else
    aDisplayInfo->UpdateMode   = EW_BSP_DISPLAY_UPDATE_PARTIAL;
  #endif

  return 1;
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayDone
*
* DESCRIPTION:
*   The function EwBspDisplayDone deinitializes the display hardware.
*
* ARGUMENTS:
*   aDisplayInfo - Display info data structure.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayDone( XDisplayInfo* aDisplayInfo )
{
  EW_UNUSED_ARG( aDisplayInfo );

  #if EW_USE_OPERATING_SYSTEM == 1

    /* destroy the LCD semaphore */
    EwBspOsSemaphoreDestroy( LcdUpdateSemaphore );

  #endif
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayGetUpdateArea
*
* DESCRIPTION:
*   The function EwBspDisplayGetUpdateArea returns the next update area
*   depending on the selected display mode:
*   In case of a synchroneous single-buffer, the function has to return the
*   the rectangular areas that correspond to the horizontal stripes (fields)
*   of the framebuffer.
*   In case of a scratch-pad buffer, the function has to return the subareas
*   that fit into the provided update rectangle.
*   During each display update, this function is called until it returns 0.
*
* ARGUMENTS:
*   aUpdateRect - Rectangular area which should be updated (redrawn).
*
* RETURN VALUE:
*   Returns 1 if a further update area can be provided, 0 otherwise.
*
*******************************************************************************/
int EwBspDisplayGetUpdateArea( XRect* aUpdateRect )
{
  #if EW_USE_DOUBLE_BUFFER == 0

    static int field = 0;

    /* check if last field is already reached */
    if ( field >= EW_NUMBER_OF_FIELDS )
    {
      field = 0;
      return 0;
    }

    /* determine rectangular area of current field */
    #if EW_SURFACE_ROTATION == 0
      *aUpdateRect = EwNewRect( 0, field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS,
        EW_FRAME_BUFFER_WIDTH, ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS );
    #endif

    #if EW_SURFACE_ROTATION == 90
      *aUpdateRect = EwNewRect( field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS, 0,
        ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS, EW_FRAME_BUFFER_WIDTH );
    #endif

    #if EW_SURFACE_ROTATION == 180
      *aUpdateRect = EwNewRect( 0, EW_FRAME_BUFFER_HEIGHT - ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS,
        EW_FRAME_BUFFER_WIDTH, EW_FRAME_BUFFER_HEIGHT - field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS );
    #endif

    #if EW_SURFACE_ROTATION == 270
      *aUpdateRect = EwNewRect( EW_FRAME_BUFFER_HEIGHT - ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS,
        0, EW_FRAME_BUFFER_HEIGHT - field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS, EW_FRAME_BUFFER_WIDTH );
    #endif

    /* next field */
    field++;

    /* sync on start line of next field to ensure save drawing operation */
    SyncOnLine( field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS );

//EwPrint( "field: %d, %4u - %4u line: %4u\n", field, aUpdateRect->Point1.Y, aUpdateRect->Point2.Y, tli_current_pos_get() & 0x0000FFFF);


    return 1;

  #endif

  return 0;
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayWaitForCompletion
*
* DESCRIPTION:
*   The function EwBspDisplayWaitForCompletion is called from the Graphics Engine
*   to ensure that all pending activities of the display system are completed, so
*   that the rendering of the next frame can start.
*   In case of a double-buffering system, the function has to wait until the
*   V-sync has occured and the pending buffer is used by the display controller.
*   In case of an external display controller, the function has to wait until
*   the transfer (update) of the graphics data has been completed and there are
*   no pending buffers.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayWaitForCompletion( void )
{
  #if EW_USE_DOUBLE_BUFFER == 1

    register uint32_t pendingBuffer = PendingFramebuffer;
    if ( CurrentFramebuffer == pendingBuffer )
      return;

    CPU_LOAD_SET_IDLE();

    #if EW_USE_OPERATING_SYSTEM == 1

      /* wait until pending framebuffer is used as current framebuffer and
         use CPU time for other tasks */
      EwBspOsSemaphoreWait( LcdUpdateSemaphore, 1000 );
      CurrentFramebuffer = pendingBuffer;

    #else

      /* wait until pending framebuffer is used as current framebuffer */
      while( CurrentFramebuffer != pendingBuffer )
        ;

    #endif

    CPU_LOAD_SET_ACTIVE();

  #endif
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayCommitBuffer
*
* DESCRIPTION:
*   The function EwBspDisplayCommitBuffer is called from the Graphics Engine
*   when the rendering of a certain buffer has been completed.
*   The type of buffer depends on the selected framebuffer concept.
*   If the display is running in a double-buffering mode, the function is called
*   after each buffer update in order to change the currently active framebuffer
*   address. Changing the framebuffer address should be synchronized with V-sync.
*   If the system is using an external graphics controller, this function is
*   responsible to start the transfer of the framebuffer content.
*
* ARGUMENTS:
*   aAddress - Address of the framebuffer to be shown on the display.
*   aX,
*   aY       - Origin of the area which has been updated by the Graphics Engine.
*   aWidth,
*   aHeight  - Size of the area which has been updated by the Graphics Engine.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayCommitBuffer( void* aAddress, int aX, int aY, int aWidth, int aHeight )
{
  #if EW_USE_DOUBLE_BUFFER == 1

    /* set pending framebuffer address to be used on next V-sync */
    PendingFramebuffer = (uint32_t)aAddress;

  #endif
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplaySetClut
*
* DESCRIPTION:
*   The function EwBspDisplaySetClut is called from the Graphics Engine
*   in order to update the hardware CLUT of the current framebuffer.
*   The function is only called when the color format of the framebuffer is
*   Index8 or LumA44.
*
* ARGUMENTS:
*   aClut - Pointer to a color lookup table with 256 entries.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplaySetClut( unsigned long* aClut )
{
  int i;
  tli_layer_lut_parameter_struct lutParams;

  tli_lut_struct_para_init( &lutParams );

  for ( i = 0; i < 256; i++ )
  {
    lutParams.layer_lut_channel_red     = (uint8_t)(*aClut >> 16);
    lutParams.layer_lut_channel_green   = (uint8_t)(*aClut >>  8);
    lutParams.layer_lut_channel_blue    = (uint8_t)(*aClut >>  0);
    lutParams.layer_table_addr          = (uint8_t)i;

    #if ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_LumA44 )
      lutParams.layer_table_addr       *= 0x10001;
    #endif

    aClut++;
    tli_lut_init( LAYER0, &lutParams );
  }

  tli_lut_enable( LAYER0 );
}


/* msy */
