/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file is part of the interface (glue layer) between an Embedded Wizard
*   generated UI application and the board support package (BSP) of a dedicated
*   target.
*   Please note: The implementation of this module is partially based on
*   examples that are provided within the GD32 Demo Suite. In case you want
*   to adapt this module to your custom specific hardware, please adapt the
*   hardware initialization code according your needs.
*   This file contains an interface to the IPA hardware of the GD32 target
*   in order to provide graphics acceleration for the Embedded Wizard generated
*   UI applications.
*   This interface is intended to be used only by the Graphics Engine of
*   Embedded Wizard.
*
*******************************************************************************/

#include "ewconfig.h"
#include <string.h>

#include "ewrte.h"
#include "ewgfx.h"
#include "ew_bsp_graphics.h"
#include "ew_bsp_clock.h"
#include "ew_bsp_os.h"

#if EW_USE_OPERATING_SYSTEM == 1

  static XSemaphoreHandle  AcceleratorSemaphoreId;

#endif


static volatile char       TransferInProgress  = 0;
static char                ConcurrentOperation = 1;

static ipa_destination_parameter_struct  ipa_destination_init_struct;
static ipa_foreground_parameter_struct   ipa_fg_init_struct;
static ipa_background_parameter_struct   ipa_bg_init_struct;


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsInit
*
* DESCRIPTION:
*   Initialize the IPA graphics accelerator.
*
* ARGUMENTS:
*   aDstColorMode - Colorformat of the destination bitmap
*
* RETURN VALUE:
*   If successful, returns != 0.
*
*******************************************************************************/
int EwBspGraphicsInit( uint32_t aDstColorMode )
{
  #if EW_USE_OPERATING_SYSTEM == 1

    /* create the accelerator semaphore */
    AcceleratorSemaphoreId = EwBspOsSemaphoreCreate( 1, 0 );

  #endif

  /* initialize parameter structs */
  ipa_foreground_struct_para_init( &ipa_fg_init_struct );
  ipa_background_struct_para_init( &ipa_bg_init_struct );
  ipa_destination_struct_para_init( &ipa_destination_init_struct );

  rcu_periph_clock_enable( RCU_IPA );

#ifdef EW_USE_IPA_INTERRUPT_MODE

  ipa_interrupt_enable( IPA_INT_FTF );
  NVIC_EnableIRQ( IPA_IRQn );
  NVIC_SetPriority( IPA_IRQn, 0x2U );

#endif

  ipa_interrupt_flag_clear( IPA_INT_FLAG_FTF );

  TransferInProgress  = 0;

  return 1;
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsDone
*
* DESCRIPTION:
*   Deinitialize the IPA graphics accelerator.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsDone( void )
{
  /* deinitialize the IPA graphics accelerator */
  ipa_deinit();

  #if EW_USE_OPERATING_SYSTEM == 1

    /* destroy the accelerator semaphore */
    EwBspOsSemaphoreDestroy( AcceleratorSemaphoreId );

  #endif
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsWaitForCompletion
*
* DESCRIPTION:
*   The function EwBspGraphicsWaitForCompletion returns as soon as the IPA has
*   completed a pending graphics instruction. If the IPA is in idle mode,
*   the function returns immediately.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsWaitForCompletion()
{
  /* return immediately if no IPA transfer is ongoing */
  if ( TransferInProgress == 0 )
    return;

  CPU_LOAD_SET_IDLE();

#ifdef EW_USE_IPA_INTERRUPT_MODE

  #if EW_USE_OPERATING_SYSTEM == 1

    EwBspOsSemaphoreWait( AcceleratorSemaphoreId, 1000 );
    TransferInProgress = 0;

  #else

    /* wait until IPA transfer is done */
    while( TransferInProgress )
      ;

  #endif

#else

  /* wait until IPA transfer is done */
  while(RESET == ipa_interrupt_flag_get( IPA_INT_FLAG_FTF ))
    ;

  ipa_interrupt_flag_clear( IPA_INT_FLAG_FTF );

  TransferInProgress = 0;

#endif

  CPU_LOAD_SET_ACTIVE();
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsConcurrentOperation
*
* DESCRIPTION:
*   The function EwBspGraphicsConcurrentOperation configures the operation mode
*   of IPA and CPU. If concurrent operation is enabled, the CPU will work in
*   parallel while the IPA is transferring data. If concurrent operation is
*   disabled, the CPU will wait everytime the IPA is active.
*   This feature is intended to limit the memory bandwidth, e.g. during display
*   update or other bandwidth consuming activities.
*
* ARGUMENTS:
*   aEnable - flag to switch on/off concurrent operation mode.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsConcurrentOperation( int aEnable )
{
  ConcurrentOperation = (char)aEnable;
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsFill
*
* DESCRIPTION:
*   The function EwBspGraphicsFill is used by the Graphics Engine, when a
*   rectangular area should be filled with a constant color by using the IPA
*   functionality.
*
* ARGUMENTS:
*   aDstAddr      - Destination address of the transfer.
*   aDstOffset    - Offset in pixel to next row within the destination.
*   aWidth,
*   aHeight       - Size of the area to fill.
*   aDstColorMode - Colorformat of the destination.
*   aSrcColor     - Color to fill the rectangular area.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsFill( uint32_t aDstAddr, uint32_t aDstOffset, uint32_t aWidth,
  uint32_t aHeight, uint32_t aDstColorMode, uint32_t aSrcColor )
{
  /* ensure that previous IPA instruction is finished */
  EwBspGraphicsWaitForCompletion();

  /* configure IPA pixel format convert mode */
  ipa_pixel_format_convert_mode_set( IPA_FILL_UP_DE );

  /* configure destination pixel format */
  ipa_destination_init_struct.destination_pf = aDstColorMode;

  /* configure destination memory base address */
  ipa_destination_init_struct.destination_memaddr = (uint32_t)aDstAddr;

  /* configure destination pre-defined alpha value RGB depending on color format */
  if ( aDstColorMode == IPA_DPF_RGB565 )
  {
    ipa_destination_init_struct.destination_preblue  = ((( aSrcColor ) >>  3 ) & 0x1F );
    ipa_destination_init_struct.destination_pregreen = ((( aSrcColor ) >> 10 ) & 0x3F );
    ipa_destination_init_struct.destination_prered   = ((( aSrcColor ) >> 19 ) & 0x1F );
    ipa_destination_init_struct.destination_prealpha = 0;
  }
  else if ( aDstColorMode == IPA_DPF_RGB888 )
  {
    ipa_destination_init_struct.destination_preblue  = ((( aSrcColor ) >>  0 ) & 0xFF );
    ipa_destination_init_struct.destination_pregreen = ((( aSrcColor ) >>  8 ) & 0xFF );
    ipa_destination_init_struct.destination_prered   = ((( aSrcColor ) >> 16 ) & 0xFF );
    ipa_destination_init_struct.destination_prealpha = 0;
  }
  else if ( aDstColorMode == IPA_DPF_ARGB4444 )
  {
    ipa_destination_init_struct.destination_preblue  = ((( aSrcColor ) >>  4 ) & 0x0F );
    ipa_destination_init_struct.destination_pregreen = ((( aSrcColor ) >> 12 ) & 0x0F );
    ipa_destination_init_struct.destination_prered   = ((( aSrcColor ) >> 20 ) & 0x0F );
    ipa_destination_init_struct.destination_prealpha = ((( aSrcColor ) >> 28 ) & 0x0F );
  }
  else /* IPA_DPF_ARGB8888 */
  {
    ipa_destination_init_struct.destination_preblue  = ((( aSrcColor ) >>  0 ) & 0xFF );
    ipa_destination_init_struct.destination_pregreen = ((( aSrcColor ) >>  8 ) & 0xFF );
    ipa_destination_init_struct.destination_prered   = ((( aSrcColor ) >> 16 ) & 0xFF );
    ipa_destination_init_struct.destination_prealpha = ((( aSrcColor ) >> 24 ) & 0xFF );
  }

  /* configure destination line offset */
  ipa_destination_init_struct.destination_lineoff = aDstOffset;

  /* configure height of the image to be processed */
  ipa_destination_init_struct.image_height = aHeight;

  /* configure width of the image to be processed */
  ipa_destination_init_struct.image_width = aWidth;

  /* IPA destination initialization */
  ipa_destination_init( &ipa_destination_init_struct );

  TransferInProgress = 1;
  ipa_transfer_enable();

  /* check for immediate completion of drawing operation */
  if ( !ConcurrentOperation )
    EwBspGraphicsWaitForCompletion();
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsFillBlend
*
* DESCRIPTION:
*   The function EwBspGraphicsFillBlend is used by the Graphics Engine, when a
*   rectangular area should be blended with a constant color by using the IPA
*   functionality.
*
* ARGUMENTS:
*   aDstAddr      - Destination address of the transfer.
*   aDstOffset    - Offset in pixel to next row within the destination.
*   aWidth,
*   aHeight       - Size of the area to blend.
*   aDstColorMode - Colorformat of the destination.
*   aSrcColor     - Color to blend the rectangular area.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsFillBlend( uint32_t aDstAddr, uint32_t aDstOffset, uint32_t aWidth,
  uint32_t aHeight, uint32_t aDstColorMode, uint32_t aSrcColor )
{
  /* ensure that previous IPA instruction is finished */
  EwBspGraphicsWaitForCompletion();

  /* configure IPA pixel format convert mode */
  ipa_pixel_format_convert_mode_set( IPA_FGBGTODE );

  /* configure destination pixel format */
  ipa_destination_init_struct.destination_pf = aDstColorMode;

  /* configure destination memory base address */
  ipa_destination_init_struct.destination_memaddr = (uint32_t)aDstAddr;

  /* configure destination pre-defined alpha value RGB */
  ipa_destination_init_struct.destination_pregreen = 0;
  ipa_destination_init_struct.destination_preblue  = 0;
  ipa_destination_init_struct.destination_prered   = 0;
  ipa_destination_init_struct.destination_prealpha = 0;

  /* configure destination line offset */
  ipa_destination_init_struct.destination_lineoff = aDstOffset;

  /* configure height of the image to be processed */
  ipa_destination_init_struct.image_height = aHeight;

  /* configure width of the image to be processed */
  ipa_destination_init_struct.image_width = aWidth;

  /* IPA destination initialization */
  ipa_destination_init( &ipa_destination_init_struct );

  /* foreground layer configuration is used as source of transfer */
  /* configure IPA foreground */
  ipa_fg_init_struct.foreground_memaddr   = (uint32_t)aDstAddr; /* avoid invalid address by setting destination address */
  ipa_fg_init_struct.foreground_lineoff   = 0;
  ipa_fg_init_struct.foreground_pf        = FOREGROUND_PPF_A8;
  ipa_fg_init_struct.foreground_prealpha  = ((( aSrcColor ) >> 24 ) & 0xFF );
  ipa_fg_init_struct.foreground_preblue   = ((( aSrcColor ) >> 0  ) & 0xFF );
  ipa_fg_init_struct.foreground_pregreen  = ((( aSrcColor ) >> 8  ) & 0xFF );
  ipa_fg_init_struct.foreground_prered    = ((( aSrcColor ) >> 16 ) & 0xFF );

  /* replace alpha channel if destination does not provide one */
  if (( aDstColorMode == BACKGROUND_PPF_ARGB8888 ) || ( aDstColorMode == BACKGROUND_PPF_RGB565 ))
    ipa_fg_init_struct.foreground_alpha_algorithm = IPA_FG_ALPHA_MODE_1;
  else
    ipa_fg_init_struct.foreground_alpha_algorithm = IPA_FG_ALPHA_MODE_2;

  ipa_foreground_init( &ipa_fg_init_struct );

  /* configure IPA background */
  ipa_bg_init_struct.background_memaddr  = aDstAddr;
  ipa_bg_init_struct.background_lineoff  = aDstOffset;
  ipa_bg_init_struct.background_pf       = aDstColorMode;
  ipa_bg_init_struct.background_alpha_algorithm = IPA_FG_ALPHA_MODE_2;
  ipa_bg_init_struct.background_prealpha = 255;
  ipa_bg_init_struct.background_preblue  = 0;
  ipa_bg_init_struct.background_pregreen = 0;
  ipa_bg_init_struct.background_prered   = 0;

  /* foreground initialization */
  ipa_background_init( &ipa_bg_init_struct );

  TransferInProgress = 1;
  ipa_transfer_enable();

  /* check for immediate completion of drawing operation */
  if ( !ConcurrentOperation )
    EwBspGraphicsWaitForCompletion();
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsCopy
*
* DESCRIPTION:
*   The function EwBspGraphicsCopy is used by the Graphics Engine, when a source
*   bitmap should be copied into a destination bitmap by using the IPA
*   functionality.
*
* ARGUMENTS:
*   aDstAddr      - Destination address of the transfer.
*   aSrcAddr      - Source address of the transfer.
*   aDstOffset    - Offset in pixel to next row within the destination bitmap.
*   aSrcOffset    - Offset in pixel to next row within the source bitmap.
*   aWidth,
*   aHeight       - Size of the area to transfer the bitmap.
*   aDstColorMode - Colorformat of the destination bitmap
*   aSrcColorMode - Colorformat of the source bitmap.
*   aSrcColor     - Color/alpha value to be used for the transfer.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsCopy( uint32_t aDstAddr, uint32_t aSrcAddr, uint32_t aDstOffset,
  uint32_t aSrcOffset, uint32_t aWidth, uint32_t aHeight, uint32_t aDstColorMode,
  uint32_t aSrcColorMode, uint32_t aSrcColor )
{
  unsigned char alpha = (unsigned char)( aSrcColor >> 24 );

  /* ensure that previous IPA instruction is finished */
  EwBspGraphicsWaitForCompletion();

  /* configure IPA pixel format convert mode */
  if ((aDstColorMode == aSrcColorMode) && (alpha == 0xFF))
    ipa_pixel_format_convert_mode_set( IPA_FGTODE );
  else
    ipa_pixel_format_convert_mode_set( IPA_FGTODE_PF_CONVERT );

  /* configure destination pixel format */
  ipa_destination_init_struct.destination_pf = aDstColorMode;

  /* configure destination memory base address */
  ipa_destination_init_struct.destination_memaddr = (uint32_t)aDstAddr;

  /* configure destination pre-defined alpha value RGB */
  ipa_destination_init_struct.destination_pregreen = 0;
  ipa_destination_init_struct.destination_preblue  = 0;
  ipa_destination_init_struct.destination_prered   = 0;
  ipa_destination_init_struct.destination_prealpha = 0;

  /* configure destination line offset */
  ipa_destination_init_struct.destination_lineoff = aDstOffset;

  /* configure height of the image to be processed */
  ipa_destination_init_struct.image_height = aHeight;

  /* configure width of the image to be processed */
  ipa_destination_init_struct.image_width = aWidth;

  /* IPA destination initialization */
  ipa_destination_init( &ipa_destination_init_struct );

  /* configure IPA foreground */
  ipa_fg_init_struct.foreground_memaddr = aSrcAddr;
  ipa_fg_init_struct.foreground_lineoff = aSrcOffset;
  ipa_fg_init_struct.foreground_pf      = aSrcColorMode;
  ipa_fg_init_struct.foreground_alpha_algorithm = IPA_FG_ALPHA_MODE_0;
  ipa_fg_init_struct.foreground_prealpha  = ((( aSrcColor ) >> 24 ) & 0xFF );

  if ( aSrcColorMode == FOREGROUND_PPF_A8 )
  {
    ipa_fg_init_struct.foreground_alpha_algorithm = IPA_FG_ALPHA_MODE_2;
    ipa_fg_init_struct.foreground_preblue   = ((( aSrcColor ) >> 0  ) & 0xFF );
    ipa_fg_init_struct.foreground_pregreen  = ((( aSrcColor ) >> 8  ) & 0xFF );
    ipa_fg_init_struct.foreground_prered    = ((( aSrcColor ) >> 16 ) & 0xFF );
  }
  else
  {
    ipa_fg_init_struct.foreground_alpha_algorithm  = (alpha == 0xFF) ? IPA_FG_ALPHA_MODE_0 : IPA_FG_ALPHA_MODE_2;
    ipa_fg_init_struct.foreground_preblue   = 0;
    ipa_fg_init_struct.foreground_pregreen  = 0;
    ipa_fg_init_struct.foreground_prered    = 0;
  }

  ipa_foreground_init( &ipa_fg_init_struct );

  TransferInProgress = 1;
  ipa_transfer_enable();

  /* check for immediate completion of drawing operation */
  if ( !ConcurrentOperation )
    EwBspGraphicsWaitForCompletion();
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsCopyBlend
*
* DESCRIPTION:
*   The function EwBspGraphicsCopyBlend is used by the Graphics Engine, when a
*   source bitmap should be blended over a destination bitmap by using the IPA
*   functionality.
*
* ARGUMENTS:
*   aDstAddr      - Destination address of the transfer.
*   aSrcAddr      - Source address of the transfer.
*   aDstOffset    - Offset in pixel to next row within the destination bitmap.
*   aSrcOffset    - Offset in pixel to next row within the source bitmap.
*   aWidth,
*   aHeight       - Size of the area to transfer the bitmap.
*   aDstColorMode - Colorformat of the destination bitmap
*   aSrcColorMode - Colorformat of the source bitmap.
*   aSrcColor     - Color/alpha value to be used for the transfer.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsCopyBlend( uint32_t aDstAddr, uint32_t aSrcAddr, uint32_t aDstOffset,
  uint32_t aSrcOffset, uint32_t aWidth, uint32_t aHeight, uint32_t aDstColorMode,
  uint32_t aSrcColorMode, uint32_t aSrcColor )
{
  unsigned char alpha = (unsigned char)( aSrcColor >> 24 );

  /* ensure that previous IPA instruction is finished */
  EwBspGraphicsWaitForCompletion();

  /* configure IPA pixel format convert mode */
  ipa_pixel_format_convert_mode_set( IPA_FGBGTODE );

  /* configure destination pixel format */
  ipa_destination_init_struct.destination_pf = aDstColorMode;

  /* configure destination memory base address */
  ipa_destination_init_struct.destination_memaddr = (uint32_t)aDstAddr;

  /* configure destination pre-defined alpha value RGB */
  ipa_destination_init_struct.destination_prered   = 0;
  ipa_destination_init_struct.destination_pregreen = 0;
  ipa_destination_init_struct.destination_preblue  = 0;
  ipa_destination_init_struct.destination_prealpha = 0;

  /* configure destination line offset */
  ipa_destination_init_struct.destination_lineoff = aDstOffset;

  /* configure height of the image to be processed */
  ipa_destination_init_struct.image_height = aHeight;

  /* configure width of the image to be processed */
  ipa_destination_init_struct.image_width = aWidth;

  /* IPA destination initialization */
  ipa_destination_init( &ipa_destination_init_struct );

  /* configure IPA foreground */
  ipa_fg_init_struct.foreground_memaddr   = aSrcAddr;
  ipa_fg_init_struct.foreground_pf        = aSrcColorMode;
  ipa_fg_init_struct.foreground_lineoff   = aSrcOffset;
  ipa_fg_init_struct.foreground_prealpha  = ((( aSrcColor ) >> 24 ) & 0xFF );

  if ( aSrcColorMode == FOREGROUND_PPF_A8 )
  {
    ipa_fg_init_struct.foreground_alpha_algorithm = IPA_FG_ALPHA_MODE_2;
    ipa_fg_init_struct.foreground_preblue   = ((( aSrcColor ) >> 0  ) & 0xFF );
    ipa_fg_init_struct.foreground_pregreen  = ((( aSrcColor ) >> 8  ) & 0xFF );
    ipa_fg_init_struct.foreground_prered    = ((( aSrcColor ) >> 16 ) & 0xFF );
  }
  else
  {
    ipa_fg_init_struct.foreground_alpha_algorithm  = (alpha == 0xFF) ? IPA_FG_ALPHA_MODE_0 : IPA_FG_ALPHA_MODE_2;
    ipa_fg_init_struct.foreground_preblue   = 0;
    ipa_fg_init_struct.foreground_pregreen  = 0;
    ipa_fg_init_struct.foreground_prered    = 0;
  }
  ipa_foreground_init( &ipa_fg_init_struct );

  /* configure IPA background */
  ipa_bg_init_struct.background_memaddr         = aDstAddr;
  ipa_bg_init_struct.background_lineoff         = aDstOffset;
  ipa_bg_init_struct.background_pf              = aDstColorMode;
  ipa_bg_init_struct.background_alpha_algorithm = IPA_BG_ALPHA_MODE_0;
  ipa_bg_init_struct.background_prealpha        = 0;
  ipa_bg_init_struct.background_preblue         = 0;
  ipa_bg_init_struct.background_pregreen        = 0;
  ipa_bg_init_struct.background_prered          = 0;
  ipa_background_init( &ipa_bg_init_struct );

  TransferInProgress = 1;
  ipa_transfer_enable();

  /* check for immediate completion of drawing operation */
  if ( !ConcurrentOperation )
    EwBspGraphicsWaitForCompletion();
}


/*******************************************************************************
* FUNCTION:
*   EwBspGraphicsLoadClut
*
* DESCRIPTION:
*   The function EwBspGraphicsLoadClut is used by the Graphics Engine, when
*   a Index8 source bitmap should be copied into a destination bitmap by using
*   the IPA functionality. In this case, the CLUT of the IPA has to be loaded
*   with the given CLUT.
*
* ARGUMENTS:
*   aClutAddr     - Address of the first color entry of the CLUT.
*   aClutSize     - Number of color entries within the CLUT.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspGraphicsLoadClut( uint32_t aClutAddr, uint32_t aClutSize )
{
  /* ensure that previous IPA instruction is finished */
  EwBspGraphicsWaitForCompletion();

  ipa_foreground_lut_init( aClutSize - 1, IPA_LUT_PF_ARGB8888, aClutAddr );

  /* check for immediate completion of CLUT loading */
  if ( !ConcurrentOperation )
    EwBspGraphicsWaitForCompletion();
}


#ifdef EW_USE_IPA_INTERRUPT_MODE

/*******************************************************************************
* FUNCTION:
*   IPA_IRQHandler
*
* DESCRIPTION:
*   IPA Interrupt Handler.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void IPA_IRQHandler( void )
{
  if ( ipa_interrupt_flag_get(IPA_INT_FLAG_FTF) != RESET )
  {
    ipa_interrupt_flag_clear( IPA_INT_FLAG_FTF );

    #if EW_USE_OPERATING_SYSTEM == 1

      EwBspOsSemaphoreRelease( AcceleratorSemaphoreId );

    #else

      TransferInProgress = 0;

    #endif
  }
}

#endif


/* msy */
